@@ -23,6 +23,7 @@ import contextlib
 import difflib
 import enum
 import errno
+import faulthandler
 import getpass
 import inspect
 import io
@@ -39,46 +40,30 @@ import subprocess
 import sys
 import tempfile
 import textwrap
+import typing
+from typing import Any, AnyStr, BinaryIO, Callable, ContextManager, IO, Iterator, List, Mapping, MutableMapping, MutableSequence, NoReturn, Optional, Sequence, Text, TextIO, Tuple, Type, Union
 import unittest
 from unittest import mock  # pylint: disable=unused-import Allow absltest.mock.
 from urllib import parse
 
-try:
-  # The faulthandler module isn't always available, and pytype doesn't
-  # understand that we're catching ImportError, so suppress the error.
-  # pytype: disable=import-error
-  import faulthandler
-  # pytype: enable=import-error
-except ImportError:
-  # We use faulthandler if it is available.
-  faulthandler = None
-
-from absl import app
+from absl import app  # pylint: disable=g-import-not-at-top
 from absl import flags
 from absl import logging
 from absl.testing import _pretty_print_reporter
 from absl.testing import xml_reporter
 
-# Make typing an optional import to avoid it being a required dependency
-# in Python 2. Type checkers will still understand the imports.
-try:
-  # pylint: disable=unused-import
-  import typing
-  from typing import Any, AnyStr, BinaryIO, Callable, ContextManager, IO, Iterator, List, Mapping, MutableMapping, MutableSequence, Optional, Sequence, Text, TextIO, Tuple, Type, Union
-  # pylint: enable=unused-import
-except ImportError:
-  pass
-else:
-  # Use an if-type-checking block to prevent leakage of type-checking only
-  # symbols. We don't want people relying on these at runtime.
-  if typing.TYPE_CHECKING:
-    # Unbounded TypeVar for general usage
-    _T = typing.TypeVar('_T')
+# Use an if-type-checking block to prevent leakage of type-checking only
+# symbols. We don't want people relying on these at runtime.
+if typing.TYPE_CHECKING:
+  # Unbounded TypeVar for general usage
+  _T = typing.TypeVar('_T')
 
-    import unittest.case
-    _OutcomeType = unittest.case._Outcome  # pytype: disable=module-attr
+  import unittest.case  # pylint: disable=g-import-not-at-top,g-bad-import-order
+
+  _OutcomeType = unittest.case._Outcome  # pytype: disable=module-attr
 
 
+# pylint: enable=g-import-not-at-top
 
 # Re-export a bunch of unittest functions we support so that people don't
 # have to import unittest to get them
@@ -103,10 +88,11 @@ __unittest = True  # pylint: disable=invalid-name
 def expectedFailureIf(condition, reason):  # pylint: disable=invalid-name
   """Expects the test to fail if the run condition is True.
 
-  Example usage:
-    @expectedFailureIf(sys.version.major == 2, "Not yet working in py2")
-    def test_foo(self):
-      ...
+  Example usage::
+
+      @expectedFailureIf(sys.version.major == 2, "Not yet working in py2")
+      def test_foo(self):
+        ...
 
   Args:
     condition: bool, whether to expect failure or not.
@@ -532,7 +518,10 @@ class _TempFile(object):
   # currently `Any` to avoid [bad-return-type] errors in the open_* methods.
   @contextlib.contextmanager
   def _open(
-      self, mode: str, encoding: str = 'utf8', errors: str = 'strict'
+      self,
+      mode: str,
+      encoding: Optional[str] = 'utf8',
+      errors: Optional[str] = 'strict',
   ) -> Iterator[Any]:
     with io.open(
         self.full_path, mode=mode, encoding=encoding, errors=errors) as fp:
@@ -589,8 +578,9 @@ class TestCase(unittest.TestCase):
   longMessage = True
 
   # Exit stacks for per-test and per-class scopes.
-  _exit_stack = None
-  _cls_exit_stack = None
+  if sys.version_info < (3, 11):
+    _exit_stack = None
+    _cls_exit_stack = None
 
   def __init__(self, *args, **kwargs):
     super(TestCase, self).__init__(*args, **kwargs)
@@ -599,17 +589,22 @@ class TestCase(unittest.TestCase):
 
   def setUp(self):
     super(TestCase, self).setUp()
-    # NOTE: Only Python 3 contextlib has ExitStack
-    if hasattr(contextlib, 'ExitStack'):
+    # NOTE: Only Python 3 contextlib has ExitStack and
+    # Python 3.11+ already has enterContext.
+    if hasattr(contextlib, 'ExitStack') and sys.version_info < (3, 11):
       self._exit_stack = contextlib.ExitStack()
       self.addCleanup(self._exit_stack.close)
 
   @classmethod
   def setUpClass(cls):
     super(TestCase, cls).setUpClass()
-    # NOTE: Only Python 3 contextlib has ExitStack and only Python 3.8+ has
-    # addClassCleanup.
-    if hasattr(contextlib, 'ExitStack') and hasattr(cls, 'addClassCleanup'):
+    # NOTE: Only Python 3 contextlib has ExitStack, only Python 3.8+ has
+    # addClassCleanup and Python 3.11+ already has enterClassContext.
+    if (
+        hasattr(contextlib, 'ExitStack')
+        and hasattr(cls, 'addClassCleanup')
+        and sys.version_info < (3, 11)
+    ):
       cls._cls_exit_stack = contextlib.ExitStack()
       cls.addClassCleanup(cls._cls_exit_stack.close)
 
@@ -623,7 +618,7 @@ class TestCase(unittest.TestCase):
     This creates a named directory on disk that is isolated to this test, and
     will be properly cleaned up by the test. This avoids several pitfalls of
     creating temporary directories for test purposes, as well as makes it easier
-    to setup directories and verify their contents. For example:
+    to setup directories and verify their contents. For example::
 
         def test_foo(self):
           out_dir = self.create_tempdir()
@@ -637,14 +632,14 @@ class TestCase(unittest.TestCase):
           self.assertTrue(os.path.exists(expected_paths[1]))
           self.assertEqual('foo', out_log.read_text())
 
-    See also: `create_tempfile()` for creating temporary files.
+    See also: :meth:`create_tempfile` for creating temporary files.
 
     Args:
       name: Optional name of the directory. If not given, a unique
         name will be generated and used.
       cleanup: Optional cleanup policy on when/if to remove the directory (and
         all its contents) at the end of the test. If None, then uses
-        `self.tempfile_cleanup`.
+        :attr:`tempfile_cleanup`.
 
     Returns:
       A _TempDir representing the created directory; see _TempDir class docs
@@ -678,7 +673,7 @@ class TestCase(unittest.TestCase):
     be properly cleaned up by the test. This avoids several pitfalls of
     creating temporary files for test purposes, as well as makes it easier
     to setup files, their data, read them back, and inspect them when
-    a test fails. For example:
+    a test fails. For example::
 
         def test_foo(self):
           output = self.create_tempfile()
@@ -690,15 +685,15 @@ class TestCase(unittest.TestCase):
     state.
     NOTE: If the file already exists, it will be made writable and overwritten.
 
-    See also: `create_tempdir()` for creating temporary directories, and
-    `_TempDir.create_file` for creating files within a temporary directory.
+    See also: :meth:`create_tempdir` for creating temporary directories, and
+    ``_TempDir.create_file`` for creating files within a temporary directory.
 
     Args:
       file_path: Optional file path for the temp file. If not given, a unique
         file name will be generated and used. Slashes are allowed in the name;
         any missing intermediate directories will be created. NOTE: This path is
         the path that will be cleaned up, including any directories in the path,
-        e.g., 'foo/bar/baz.txt' will `rm -r foo`.
+        e.g., ``'foo/bar/baz.txt'`` will ``rm -r foo``.
       content: Optional string or
         bytes to initially write to the file. If not
         specified, then an empty file is created.
@@ -710,7 +705,7 @@ class TestCase(unittest.TestCase):
         `content` is text.
       cleanup: Optional cleanup policy on when/if to remove the directory (and
         all its contents) at the end of the test. If None, then uses
-        `self.tempfile_cleanup`.
+        :attr:`tempfile_cleanup`.
 
     Returns:
       A _TempFile representing the created file; see _TempFile class docs for
@@ -734,7 +729,7 @@ class TestCase(unittest.TestCase):
     (e.g. `TestCase.enter_context`), the context is exited after the test
     class's tearDownClass call.
 
-    Contexts are are exited in the reverse order of entering. They will always
+    Contexts are exited in the reverse order of entering. They will always
     be exited, regardless of test failure/success.
 
     This is useful to eliminate per-test boilerplate when context managers
@@ -748,6 +743,9 @@ class TestCase(unittest.TestCase):
     Args:
       manager: The context manager to enter.
     """
+    if sys.version_info >= (3, 11):
+      return self.enterContext(manager)
+
     if not self._exit_stack:
       raise AssertionError(
           'self._exit_stack is not set: enter_context is Py3-only; also make '
@@ -757,6 +755,9 @@ class TestCase(unittest.TestCase):
   @enter_context.classmethod
   def enter_context(cls, manager):  # pylint: disable=no-self-argument
     # type: (ContextManager[_T]) -> _T
+    if sys.version_info >= (3, 11):
+      return cls.enterClassContext(manager)
+
     if not cls._cls_exit_stack:
       raise AssertionError(
           'cls._cls_exit_stack is not set: cls.enter_context requires '
@@ -799,6 +800,7 @@ class TestCase(unittest.TestCase):
   ) -> None:
     """Adds `function` as cleanup when the test case succeeds."""
     outcome = self._outcome
+    assert outcome is not None
     previous_failure_count = (
         len(outcome.result.failures)
         + len(outcome.result.errors)
@@ -818,6 +820,7 @@ class TestCase(unittest.TestCase):
     """Returns whether test is passed. Expected to be called during cleanup."""
     outcome = self._outcome
     if sys.version_info[:2] >= (3, 11):
+      assert outcome is not None
       current_failure_count = (
           len(outcome.result.failures)
           + len(outcome.result.errors)
@@ -926,6 +929,12 @@ class TestCase(unittest.TestCase):
       prefix = [prefix]
       prefix_len = 1
 
+    if isinstance(whole, abc.Mapping) or isinstance(whole, abc.Set):
+      self.fail(
+          'For whole: Mapping or Set objects are not supported, found type: %s'
+          % type(whole),
+          msg,
+      )
     try:
       whole_len = len(whole)
     except (TypeError, NotImplementedError):
@@ -1075,10 +1084,10 @@ class TestCase(unittest.TestCase):
     """Asserts that two sequences have the same elements (in any order).
 
     This method, unlike assertCountEqual, doesn't care about any
-    duplicates in the expected and actual sequences.
+    duplicates in the expected and actual sequences::
 
-      >> assertSameElements([1, 1, 1, 0, 0, 0], [0, 1])
-      # Doesn't raise an AssertionError
+        # Doesn't raise an AssertionError
+        assertSameElements([1, 1, 1, 0, 0, 0], [0, 1])
 
     If possible, you should use assertCountEqual instead of
     assertSameElements.
@@ -1174,6 +1183,7 @@ class TestCase(unittest.TestCase):
     expression (a string or re compiled object) instead of a list.
 
     Notes:
+
     1. This function uses substring matching, i.e. the matching
        succeeds if *any* substring of the error message matches *any*
        regex in the list.  This is more convenient for the user than
@@ -1334,6 +1344,8 @@ class TestCase(unittest.TestCase):
       if not issubclass(exc_type, self.expected_exception):
         return False
       self.test_func(exc_value)
+      if exc_value:
+        self.exception = exc_value.with_traceback(None)
       return True
 
   @typing.overload
@@ -1523,39 +1535,39 @@ class TestCase(unittest.TestCase):
     """Asserts that total ordering has been implemented correctly.
 
     For example, say you have a class A that compares only on its attribute x.
-    Comparators other than __lt__ are omitted for brevity.
+    Comparators other than ``__lt__`` are omitted for brevity::
 
-    class A(object):
-      def __init__(self, x, y):
-        self.x = x
-        self.y = y
+        class A(object):
+          def __init__(self, x, y):
+            self.x = x
+            self.y = y
 
-      def __hash__(self):
-        return hash(self.x)
+          def __hash__(self):
+            return hash(self.x)
 
-      def __lt__(self, other):
-        try:
-          return self.x < other.x
-        except AttributeError:
-          return NotImplemented
+          def __lt__(self, other):
+            try:
+              return self.x < other.x
+            except AttributeError:
+              return NotImplemented
 
     assertTotallyOrdered will check that instances can be ordered correctly.
-    For example,
+    For example::
 
-    self.assertTotallyOrdered(
-      [None],  # None should come before everything else.
-      [1],     # Integers sort earlier.
-      [A(1, 'a')],
-      [A(2, 'b')],  # 2 is after 1.
-      [A(3, 'c'), A(3, 'd')],  # The second argument is irrelevant.
-      [A(4, 'z')],
-      ['foo'])  # Strings sort last.
+        self.assertTotallyOrdered(
+            [None],  # None should come before everything else.
+            [1],  # Integers sort earlier.
+            [A(1, 'a')],
+            [A(2, 'b')],  # 2 is after 1.
+            [A(3, 'c'), A(3, 'd')],  # The second argument is irrelevant.
+            [A(4, 'z')],
+            ['foo'])  # Strings sort last.
 
     Args:
-     *groups: A list of groups of elements.  Each group of elements is a list
-         of objects that are equal.  The elements in each group must be less
-         than the elements in the group after it.  For example, these groups are
-         totally ordered: [None], [1], [2, 2], [3].
+      *groups: A list of groups of elements.  Each group of elements is a list
+        of objects that are equal.  The elements in each group must be less
+        than the elements in the group after it.  For example, these groups are
+        totally ordered: ``[None]``, ``[1]``, ``[2, 2]``, ``[3]``.
       **kwargs: optional msg keyword argument can be passed.
     """
 
@@ -1758,7 +1770,8 @@ class TestCase(unittest.TestCase):
     # rather than just stopping at the first
     problems = []
 
-    _walk_structure_for_problems(a, b, aname, bname, problems)
+    _walk_structure_for_problems(a, b, aname, bname, problems,
+                                 self.assertEqual, self.failureException)
 
     # Avoid spamming the user toooo much
     if self.maxDiff is not None:
@@ -1811,9 +1824,9 @@ class TestCase(unittest.TestCase):
 
     return super(TestCase, self)._getAssertEqualityFunc(first, second)
 
-  def fail(self, msg=None, prefix=None):
-    """Fail immediately with the given message, optionally prefixed."""
-    return super(TestCase, self).fail(self._formatMessage(prefix, msg))
+  def fail(self, msg=None, user_msg=None) -> NoReturn:
+    """Fail immediately with the given standard message and user message."""
+    return super(TestCase, self).fail(self._formatMessage(user_msg, msg))
 
 
 def _sorted_list_difference(expected, actual):
@@ -1889,7 +1902,9 @@ def _are_both_of_mapping_type(a, b):
       b, abc.Mapping)
 
 
-def _walk_structure_for_problems(a, b, aname, bname, problem_list):
+def _walk_structure_for_problems(
+    a, b, aname, bname, problem_list, leaf_assert_equal_func, failure_exception
+):
   """The recursive comparison behind assertSameStructure."""
   if type(a) != type(b) and not (  # pylint: disable=unidiomatic-typecheck
       _are_both_of_integer_type(a, b) or _are_both_of_sequence_type(a, b) or
@@ -1917,7 +1932,7 @@ def _walk_structure_for_problems(a, b, aname, bname, problem_list):
       if k in b:
         _walk_structure_for_problems(
             a[k], b[k], '%s[%r]' % (aname, k), '%s[%r]' % (bname, k),
-            problem_list)
+            problem_list, leaf_assert_equal_func, failure_exception)
       else:
         problem_list.append(
             "%s has [%r] with value %r but it's missing in %s" %
@@ -1935,7 +1950,7 @@ def _walk_structure_for_problems(a, b, aname, bname, problem_list):
     for i in range(minlen):
       _walk_structure_for_problems(
           a[i], b[i], '%s[%d]' % (aname, i), '%s[%d]' % (bname, i),
-          problem_list)
+          problem_list, leaf_assert_equal_func, failure_exception)
     for i in range(minlen, len(a)):
       problem_list.append('%s has [%i] with value %r but %s does not' %
                           (aname, i, a[i], bname))
@@ -1944,7 +1959,9 @@ def _walk_structure_for_problems(a, b, aname, bname, problem_list):
                           (aname, i, bname, b[i]))
 
   else:
-    if a != b:
+    try:
+      leaf_assert_equal_func(a, b)
+    except failure_exception:
       problem_list.append('%s is %r but %s is %r' % (aname, a, bname, b))
 
 
@@ -2040,12 +2057,14 @@ def main(*args, **kwargs):
 
   Usually this function is called without arguments, so the
   unittest.TestProgram instance will get created with the default settings,
-  so it will run all test methods of all TestCase classes in the __main__
+  so it will run all test methods of all TestCase classes in the ``__main__``
   module.
 
   Args:
-    *args: Positional arguments passed through to unittest.TestProgram.__init__.
-    **kwargs: Keyword arguments passed through to unittest.TestProgram.__init__.
+    *args: Positional arguments passed through to
+        ``unittest.TestProgram.__init__``.
+    **kwargs: Keyword arguments passed through to
+        ``unittest.TestProgram.__init__``.
   """
   print_python_version()
   _run_in_app(run_tests, args, kwargs)
@@ -2065,7 +2084,7 @@ def _is_in_app_main():
 def _register_sigterm_with_faulthandler():
   # type: () -> None
   """Have faulthandler dump stacks on SIGTERM.  Useful to diagnose timeouts."""
-  if faulthandler and getattr(faulthandler, 'register', None):
+  if getattr(faulthandler, 'register', None):
     # faulthandler.register is not available on Windows.
     # faulthandler.enable() is already called by app.run.
     try:
@@ -2178,29 +2197,29 @@ def skipThisClass(reason):
   implementations between a number of concrete testcase classes.
 
   Example usage, showing how you can share some common test methods between
-  subclasses. In this example, only 'BaseTest' will be marked as skipped, and
-  not RealTest or SecondRealTest:
+  subclasses. In this example, only ``BaseTest`` will be marked as skipped, and
+  not RealTest or SecondRealTest::
 
-    @absltest.skipThisClass("Shared functionality")
-    class BaseTest(absltest.TestCase):
-      def test_simple_functionality(self):
-        self.assertEqual(self.system_under_test.method(), 1)
+      @absltest.skipThisClass("Shared functionality")
+      class BaseTest(absltest.TestCase):
+        def test_simple_functionality(self):
+          self.assertEqual(self.system_under_test.method(), 1)
 
-    class RealTest(BaseTest):
-      def setUp(self):
-        super().setUp()
-        self.system_under_test = MakeSystem(argument)
+      class RealTest(BaseTest):
+        def setUp(self):
+          super().setUp()
+          self.system_under_test = MakeSystem(argument)
 
-      def test_specific_behavior(self):
-        ...
+        def test_specific_behavior(self):
+          ...
 
-    class SecondRealTest(BaseTest):
-      def setUp(self):
-        super().setUp()
-        self.system_under_test = MakeSystem(other_arguments)
+      class SecondRealTest(BaseTest):
+        def setUp(self):
+          super().setUp()
+          self.system_under_test = MakeSystem(other_arguments)
 
-      def test_other_behavior(self):
-        ...
+        def test_other_behavior(self):
+          ...
 
   Args:
     reason: The reason we have a skip in place. For instance: 'shared test
@@ -2275,7 +2294,7 @@ class TestLoader(unittest.TestLoader):
     for name in dir(testCaseClass):
       if _is_suspicious_attribute(testCaseClass, name):
         raise TypeError(TestLoader._ERROR_MSG % name)
-    names = super(TestLoader, self).getTestCaseNames(testCaseClass)
+    names = list(super(TestLoader, self).getTestCaseNames(testCaseClass))
     if self._randomize_ordering_seed is not None:
       logging.info(
           'Randomizing test order with seed: %d', self._randomize_ordering_seed)
@@ -2298,8 +2317,7 @@ def get_default_xml_output_filename():
         os.path.splitext(os.path.basename(sys.argv[0]))[0] + '.xml')
 
 
-def _setup_filtering(argv):
-  # type: (MutableSequence[Text]) -> None
+def _setup_filtering(argv: MutableSequence[str]) -> bool:
   """Implements the bazel test filtering protocol.
 
   The following environment variable is used in this method:
@@ -2314,16 +2332,20 @@ def _setup_filtering(argv):
 
   Args:
     argv: the argv to mutate in-place.
+
+  Returns:
+    Whether test filtering is requested.
   """
   test_filter = os.environ.get('TESTBRIDGE_TEST_ONLY')
   if argv is None or not test_filter:
-    return
+    return False
 
   filters = shlex.split(test_filter)
   if sys.version_info[:2] >= (3, 7):
     filters = ['-k=' + test_filter for test_filter in filters]
 
   argv[1:1] = filters
+  return True
 
 
 def _setup_test_runner_fail_fast(argv):
@@ -2350,8 +2372,9 @@ def _setup_test_runner_fail_fast(argv):
   argv[1:1] = ['--failfast']
 
 
-def _setup_sharding(custom_loader=None):
-  # type: (Optional[unittest.TestLoader]) -> unittest.TestLoader
+def _setup_sharding(
+    custom_loader: Optional[unittest.TestLoader] = None,
+) -> Tuple[unittest.TestLoader, Optional[int]]:
   """Implements the bazel sharding protocol.
 
   The following environment variables are used in this method:
@@ -2370,8 +2393,10 @@ def _setup_sharding(custom_loader=None):
     custom_loader: A TestLoader to be made sharded.
 
   Returns:
-    The test loader for shard-filtering or the standard test loader, depending
-    on the sharding environment variables.
+    A tuple of ``(test_loader, shard_index)``. ``test_loader`` is for
+    shard-filtering or the standard test loader depending on the sharding
+    environment variables. ``shard_index`` is the shard index, or ``None`` when
+    sharding is not used.
   """
 
   # It may be useful to write the shard file even if the other sharding
@@ -2389,7 +2414,7 @@ def _setup_sharding(custom_loader=None):
   base_loader = custom_loader or TestLoader()
   if 'TEST_TOTAL_SHARDS' not in os.environ:
     # Not using sharding, use the expected test loader.
-    return base_loader
+    return base_loader, None
 
   total_shards = int(os.environ['TEST_TOTAL_SHARDS'])
   shard_index = int(os.environ['TEST_SHARD_INDEX'])
@@ -2418,25 +2443,70 @@ def _setup_sharding(custom_loader=None):
     return [x for x in ordered_names if x in filtered_names]
 
   base_loader.getTestCaseNames = getShardedTestCaseNames
-  return base_loader
+  return base_loader, shard_index
 
 
-# pylint: disable=line-too-long
-def _run_and_get_tests_result(argv, args, kwargs, xml_test_runner_class):
-  # type: (MutableSequence[Text], Sequence[Any], MutableMapping[Text, Any], Type) -> unittest.TestResult
-  # pylint: enable=line-too-long
-  """Same as run_tests, except it returns the result instead of exiting."""
+def _run_and_get_tests_result(
+    argv: MutableSequence[str],
+    args: Sequence[Any],
+    kwargs: MutableMapping[str, Any],
+    xml_test_runner_class: Type[unittest.TextTestRunner],
+) -> Tuple[unittest.TestResult, bool]:
+  """Same as run_tests, but it doesn't exit.
+
+  Args:
+    argv: sys.argv with the command-line flags removed from the front, i.e. the
+      argv with which :func:`app.run()<absl.app.run>` has called
+      ``__main__.main``. It is passed to
+      ``unittest.TestProgram.__init__(argv=)``, which does its own flag parsing.
+      It is ignored if kwargs contains an argv entry.
+    args: Positional arguments passed through to
+      ``unittest.TestProgram.__init__``.
+    kwargs: Keyword arguments passed through to
+      ``unittest.TestProgram.__init__``.
+    xml_test_runner_class: The type of the test runner class.
+
+  Returns:
+    A tuple of ``(test_result, fail_when_no_tests_ran)``.
+    ``fail_when_no_tests_ran`` indicates whether the test should fail when
+    no tests ran.
+  """
 
   # The entry from kwargs overrides argv.
   argv = kwargs.pop('argv', argv)
 
+  if sys.version_info[:2] >= (3, 12):
+    # Python 3.12 unittest changed the behavior from PASS to FAIL in
+    # https://github.com/python/cpython/pull/102051. absltest follows this.
+    fail_when_no_tests_ran = True
+  else:
+    # Historically, absltest and unittest before Python 3.12 passes if no tests
+    # ran.
+    fail_when_no_tests_ran = False
+
   # Set up test filtering if requested in environment.
-  _setup_filtering(argv)
+  if _setup_filtering(argv):
+    # When test filtering is requested, ideally we also want to fail when no
+    # tests ran. However, the test filters are usually done when running bazel.
+    # When you run multiple targets, e.g. `bazel test //my_dir/...
+    # --test_filter=MyTest`, you don't necessarily want individual tests to fail
+    # because no tests match in that particular target.
+    # Due to this use case, we don't fail when test filtering is requested via
+    # the environment variable from bazel.
+    fail_when_no_tests_ran = False
+
   # Set up --failfast as requested in environment
   _setup_test_runner_fail_fast(argv)
 
   # Shard the (default or custom) loader if sharding is turned on.
-  kwargs['testLoader'] = _setup_sharding(kwargs.get('testLoader', None))
+  kwargs['testLoader'], shard_index = _setup_sharding(
+      kwargs.get('testLoader', None)
+  )
+  if shard_index is not None and shard_index > 0:
+    # When sharding is requested, all the shards except the first one shall not
+    # fail when no tests ran. This happens when the shard count is greater than
+    # the test case count.
+    fail_when_no_tests_ran = False
 
   # XML file name is based upon (sorted by priority):
   # --xml_output_file flag, XML_OUTPUT_FILE variable,
@@ -2514,9 +2584,13 @@ def _run_and_get_tests_result(argv, args, kwargs, xml_test_runner_class):
   # on argv, which is sys.argv without the command-line flags.
   kwargs['argv'] = argv
 
+  # Request unittest.TestProgram to not exit. The exit will be handled by
+  # `absltest.run_tests`.
+  kwargs['exit'] = False
+
   try:
     test_program = unittest.TestProgram(*args, **kwargs)
-    return test_program.result
+    return test_program.result, fail_when_no_tests_ran
   finally:
     if xml_buffer:
       try:
@@ -2526,9 +2600,11 @@ def _run_and_get_tests_result(argv, args, kwargs, xml_test_runner_class):
         xml_buffer.close()
 
 
-def run_tests(argv, args, kwargs):  # pylint: disable=line-too-long
-  # type: (MutableSequence[Text], Sequence[Any], MutableMapping[Text, Any]) -> None
-  # pylint: enable=line-too-long
+def run_tests(
+    argv: MutableSequence[Text],
+    args: Sequence[Any],
+    kwargs: MutableMapping[Text, Any],
+) -> None:
   """Executes a set of Python unit tests.
 
   Most users should call absltest.main() instead of run_tests.
@@ -2540,14 +2616,22 @@ def run_tests(argv, args, kwargs):  # pylint: disable=line-too-long
 
   Args:
     argv: sys.argv with the command-line flags removed from the front, i.e. the
-      argv with which app.run() has called __main__.main. It is passed to
-      unittest.TestProgram.__init__(argv=), which does its own flag parsing. It
-      is ignored if kwargs contains an argv entry.
-    args: Positional arguments passed through to unittest.TestProgram.__init__.
-    kwargs: Keyword arguments passed through to unittest.TestProgram.__init__.
+      argv with which :func:`app.run()<absl.app.run>` has called
+      ``__main__.main``. It is passed to
+      ``unittest.TestProgram.__init__(argv=)``, which does its own flag parsing.
+      It is ignored if kwargs contains an argv entry.
+    args: Positional arguments passed through to
+      ``unittest.TestProgram.__init__``.
+    kwargs: Keyword arguments passed through to
+      ``unittest.TestProgram.__init__``.
   """
-  result = _run_and_get_tests_result(
-      argv, args, kwargs, xml_reporter.TextAndXMLTestRunner)
+  result, fail_when_no_tests_ran = _run_and_get_tests_result(
+      argv, args, kwargs, xml_reporter.TextAndXMLTestRunner
+  )
+  if fail_when_no_tests_ran and result.testsRun == 0:
+    # Python 3.12 unittest exits with 5 when no tests ran. The code comes from
+    # pytest which does the same thing.
+    sys.exit(5)
   sys.exit(not result.wasSuccessful())
 
 
